/*
 *  Citrusleaf Aerospike Java Library
 *
 *  Copyright 2009-2010 by Citrusleaf, Inc.  All rights reserved.
 *  
 *  Availability of this source code to partners and customers includes
 *  redistribution rights covered by individual contract. Please check
 *  your contract for exact rights and responsibilities.
 */
package net.citrusleaf;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import net.citrusleaf.CitrusleafClient.ClResultCode;


public class CLConnectionManager implements Runnable{

	public static final int n_server_partitions = 4096;
	
	protected HashSet<Host> hosts;
	protected int directoryGenerationCount;
	protected Thread theThread;
	public int id; //useful for tracking how many we have
	private	boolean	dying;
	
	private HashMap<String, CLNode> nodesHashMap;  // String is the node name - 
	public ArrayList<CLNode> nodesArray;          // nodes, as an ordered list.
	//public ConcurrentHashMap<String, CLNode> partitionReadHashMap;
	public ConcurrentHashMap<String, CLNode> partitionWriteHashMap;
	private volatile int errorCount = 0;
		
	public enum PartitionType {
		READ,
		WRITE
	}
	
	public CLConnectionManager(String hostname, int port, int timeoutMillis) throws Exception
	{				
		hosts = new HashSet<Host>();
		directoryGenerationCount = 0;
		theThread = null;
		dying = false;	
		
		nodesHashMap = new HashMap<String, CLNode>();  // nodes, as a hashmap
		nodesArray = new ArrayList<CLNode>();        // nodes, as an ordered list.
		//partitionReadHashMap = new ConcurrentHashMap<String, CLNode>(1024 * 8); // hints for best node for a partition
		partitionWriteHashMap = new ConcurrentHashMap<String, CLNode>(1024 * 8); // hints for best node for a partition
		
        // Add initial host.
        if (addHost(hostname, port) != ClResultCode.OK) {
            throw new Exception("Failed to add host " + hostname + ':' + port + " to cluster");
        }
        
        waitTillStabilized(timeoutMillis);

		theThread = new Thread(this);
        theThread.setDaemon(true);
		theThread.start();
	}
	
	public void close() {
		dying = true;	
	}

	public ClResultCode addHost(String hostname, int port) {
		String nodeName = CitrusleafInfo.get(hostname, port, "node");
		
		if (nodeName == null) {
			return ClResultCode.SERVER_NOT_AVAILABLE;
		}

		Host host = new Host(hostname, port);
		synchronized (hosts) {
			hosts.add(host);
		}
		return ClResultCode.OK;
	}
	
    /**
     * Tend the cluster until it has stabilized and return control.
     * This helps avoid initial database request timeout issues when
     * a large number of threads are initiated at client startup.
     * 
     * If the cluster has not stabilized by the timeout, return
     * control as well.  Do not return an error since future 
     * database requests may still succeed.
     */
    private void waitTillStabilized(int timeoutMillis) {
		long limit = System.currentTimeMillis() + timeoutMillis;
		int count = nodesArray.size();
		
		do {
			tend();
		
			// Check to see if cluster has changed since the last Tend().
			// If not, assume cluster has stabilized and return.
			if (count == nodesArray.size())
				return;
			
			// Sleep for 2 seconds. This area could use more smarts, but how to tell if we're 
			// getting overwhelmed with small requests, or large requests.
			try {
				Thread.sleep(1);
			} 
			catch (InterruptedException e) {
			}
			
			count = nodesArray.size();
		} while (System.currentTimeMillis() < limit);
    }

    // at the moment, we don't do anything fancy to match the request key and 
	// the node.  We just increment the node index.
	private int curIdx = 0;

	public CLConnection getConnection(String ns, CLNode node, int timeout)
	{
		CLConnection conn = null;
		try {
			conn = node.getConnection();
			if (conn != null) {
				if (conn.isConnected() == false) {
					conn.connect(timeout);
				}
				if (conn.isConnected() == true) {
					return(conn);
				}								
				conn.poolReturn(false);
				conn = null;
			}
		} catch (Exception e) {
			if (CitrusleafClient.infoEnabled()) {
				CitrusleafClient.info("connection manager: could not get connection: " + e.toString() );
			}
		}

		return conn;
	}

	public CLConnection getConnection(String ns, int partition_id, PartitionType partition_type, int timeout) throws IOException
	{
		CLConnection conn = null;
		long deadline = timeout > 0 ? (System.currentTimeMillis() + timeout) : 0;
		
		// see if we have a specific recommendation for this partition
		String ns_part = String.format("%s:%d",ns, partition_id);
		CLNode node = partitionWriteHashMap.get(ns_part);
		 
		if (node != null) {				
			try {
				conn = node.getConnection();
			} catch (Exception e) {
				if (CitrusleafClient.debugEnabled()) {
					CitrusleafClient.debug("connection manager: could not get connection: " + e.toString() );
				}
			}

			if (conn != null) {
				if (conn.isConnected() == false) {
					conn.connect(timeout);
				}
				if (conn.isConnected() == true) 
					return(conn);
								
				conn.poolReturn(false);
				conn = null;
			}
		}
		
		// get a "random" - round-robin - node
		int myTries = -1;

		do {
			if (conn != null) {
				if (CitrusleafClient.debugEnabled()) {		
					CitrusleafClient.debug(" internal error: in the random section with a connection, dunno why "+myTries);
				}
			}
			conn = null;
		
			synchronized(nodesArray) {
				
				int listSize = nodesArray.size();
				if ( listSize ==  0 )
					return null;
				if (myTries == -1) myTries = listSize;
				
				if ( curIdx >= listSize ) {
					curIdx = 0;
				}
				
				try{
	//				CitrusleafClient.info(" getting connection at " + curIdx);
					node = nodesArray.get(curIdx);
					conn = node.getConnection();

					curIdx++;
					
				} catch( Exception e ) {
					if (CitrusleafClient.infoEnabled()) {
						CitrusleafClient.info(" ConnectionManager: get connection: Caught exception " + e);
					}
					if (conn != null) {
						conn.poolReturn(false);
						conn = null;
					}
				}				
			}
			
			if (conn != null) {
				if (conn.isConnected() == true) 
					return(conn);
				
				if (deadline != 0) { 
					int c_timeout = (int) (deadline - System.currentTimeMillis());
					if (c_timeout>0) {
						conn.connect(c_timeout);						
					} // I get here because "lots" of time has passed between when deadline was set and now. (happens with short timeouts)
				}
				else {
					conn.connect();
				}
				
				if (conn.isConnected() == true)
					return(conn);
				
				conn.poolReturn(false); // not connected
				conn = null;					
			}
		
			if (deadline != 0) {
				if (deadline < System.currentTimeMillis()) {
					throw new IOException("connection to cluster timed out");
				}
			}
			
			myTries--;
			
		} while( myTries > 0 );

		if( conn == null ) {
			throw new IOException("can't connect to cluster");
		}
		
		return conn;
	}
	
	public void releaseConnection(CLConnection conn, boolean connected)
	{
		conn.poolReturn(connected);
	}
	
	public void closeAllConnections()
	{
		synchronized(nodesArray){
			Iterator<CLNode> i = nodesArray.iterator();
			while( i.hasNext() ){
				CLNode node = i.next();
				node.closeAllConnections();
			}
		}	
	}
	
	//
	// check through the tables and remove this node - life has changed
	//
	public void unassociateNodePartitionAll(CLNode node, PartitionType partition_type)
	{
		ConcurrentHashMap<String,CLNode> pMap = null;
		
		if (partition_type == PartitionType.WRITE) {
			pMap = partitionWriteHashMap;
		}
		//else if (partition_type == PartitionType.READ) 
		//	pMap = partitionReadHashMap;
		else {
			CitrusleafClient.error("Invalid partition type: " + partition_type);
			return;
		}
		ArrayList<String> deleteList = new ArrayList<String>();

		for (Map.Entry<String, CLNode> entry : pMap.entrySet()) {
			if (entry.getValue() == node) {
				deleteList.add(entry.getKey());
			}
		}
		
		for (String delete : deleteList) {
			pMap.remove(delete);
		}		
	}

	public void unassociateNodePartitionAll(CLNode node)
	{
		unassociateNodePartitionAll(node, PartitionType.WRITE);
		//unassociateNodePartitionAll(node, PartitionType.READ);		
	}
	
	public void associateNodePartition(CLNode node, String ns_part, PartitionType partition_type)
	{
		String namespace = null;
		int partitionId = 0;
		
		try {
			// Key format: <namespace>:<partition id>
			String[] list = ns_part.split(":");
			
			if (list.length != 2)
				throw new Exception();
			
			namespace = list[0].trim();
			if (namespace.length() <= 0 || namespace.length() >= 32)
				throw new Exception();
				
			partitionId = Integer.parseInt(list[1]);
			if (partitionId < 0 || partitionId >= n_server_partitions)
				throw new Exception();
		}
		catch (Exception ex) {
			errorCount++;
			
			// Avoid spamming error logs.
			if (errorCount < 1000 || errorCount % 10000 == 0) {
				String key = (ns_part.length() <= 100)? ns_part : ns_part.substring(0, 100);
				CitrusleafClient.error("Invalid " + partition_type + " partition hashmap key. length=" + 
					ns_part.length() + " value=" + key);
			}
			return;
		}
		
		ConcurrentHashMap<String,CLNode> pMap = null;
		
		if (partition_type == PartitionType.WRITE) {
			pMap = partitionWriteHashMap;
		}
		//else if (partition_type == PartitionType.READ) 
		//	pMap = partitionReadHashMap;
		else {
			CitrusleafClient.error("Invalid partition type: " + partition_type);
			return;
		}
		
		String key = String.format("%s:%d", namespace, partitionId);
		pMap.put(key, node);
		
		if (pMap.size() > 100000) {
			errorCount++;
			
			// Avoid spamming error logs.
			if (errorCount < 1000 || errorCount % 10000 == 0) {
				CitrusleafClient.error("Invalid " + partition_type + " partition hashmap size: " +  pMap.size());
			}
		}
	}

	// When you set the name of a node, you need to set the association
	// at the manager layer.
	public boolean associateNodeName(CLNode node, String name)
	{
		synchronized(nodesArray) {
			CLNode existingNode = nodesHashMap.get(name);
			if (existingNode == null) {
				nodesArray.add(node);
				nodesHashMap.put(name, node);
				return(true);
			}
			else {
				existingNode.addAliases(node); // merge in the alias list
			}
		}
//		CitrusleafClient.info(" connection manager attempt to associate "+name+" failed, already exists");
		return(false);
	}
	
	
	// has the cluster formed its first connections?
	public Boolean connected()
	{
		if (dying == true) return false;
		return !nodesArray.isEmpty();
	}
	
	
	public List<String> getNodeNameList()
	{
		List<String> nodeNameList = new ArrayList<String>();
		synchronized(nodesArray){
			for (CLNode node : nodesArray) {
				nodeNameList.add(node.name);
			}
		}
		return nodeNameList;		
	}
	
	public CLNode getNodeFromNodeName(String node_name)
	{
		synchronized(nodesHashMap){
			return nodesHashMap.get(node_name);		
		}
	}
	
	public void run() {
		try {
			while (dying == false) {
				tend();
	
				// Sleep for 2 seconds. This area could use more smarts, but how to tell if we're 
				// getting overwhelmed with small requests, or large requests.
				try {
					Thread.sleep(2000);	
				} 
				catch (InterruptedException e) {
				}	
			}
		}
		catch (Exception ex) {
			CitrusleafClient.error("Cluster tend aborted: " + ex.getMessage());
		}
	}
	
    /**
     * Check health of all nodes in the cluster.
     */
	public void tend() {

		// CitrusleafClient.info(" connection manager run: "+System.currentTimeMillis() );
		ArrayList<CLNode> nodesCopy = new ArrayList<CLNode>();
		ArrayList<CLNode> possibleFriends = new ArrayList<CLNode>();
		
		// if the host array is empty, use the seed
		if (nodesArray.isEmpty()) {
			// CitrusleafClient.info("nodes array empty: adding the seed nodes to the nodes copy");
			synchronized (hosts) {
				for (Host host : hosts) {
					CLNode node = new CLNode(this, host.hostname, host.port);
					nodesCopy.add(node);
				}
			}
		}
		else {		
			// clone the array so I can iterate over the blocking calls
			synchronized(nodesArray) {
				for (CLNode node : nodesArray) {
					nodesCopy.add(node);
				}
			}
		}

		try {			
			// Ask each node for connection information. 
			// This will populate the possible friends array
			for (CLNode node : nodesCopy) {
				// CitrusleafClient.info(" connection manager: first pass refreshing the node info: "+node.name);
				node.refreshNodeInfo(possibleFriends, true);
			}

			// If we have possible friends, populate their info before comparing
			// but don't get *THEIR* friends, that creates too much recursion, 
			// catch those friends the next time around
			//
			// don't want the part information or friends list this time
			// SIDE EFFECT - will add to the various node list structures if its new
			// or forget it if its not
			for (CLNode node : possibleFriends) {
				// CitrusleafClient.info(" connection manager: refresh node info: "+node.name);
				node.refreshNodeInfo(null, false); 
			}
		} 
		catch (Exception e) {
			CitrusleafClient.warn("Failed to refresh nodes: " + e.getMessage());
		}
	}
		
	public void dunNode(CLNode node) {
		if (CitrusleafClient.infoEnabled()) {
			CitrusleafClient.info(" DUNNING node "+node.name);
		}
		
		synchronized(nodesArray){
			nodesHashMap.remove(node.name);
			nodesArray.remove(node);
			unassociateNodePartitionAll(node);
		}
	}		
	
	private static class Host {
		String hostname;
		int port;
		
		public Host(String hostname, int port) {
			this.hostname = hostname;
			this.port = port;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = prime + hostname.hashCode();
			result = prime * result + port;
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			Host other = (Host) obj;
			return this.hostname.equals(other.hostname) && this.port == other.port;
		}
	}
}
