/*
 *  Citrusleaf Aerospike Java Library
 *
 *  Copyright 2009-2010 by Citrusleaf, Inc.  All rights reserved.
 *  
 *  Availability of this source code to partners and customers includes
 *  redistribution rights covered by individual contract. Please check
 *  your contract for exact rights and responsibilities.
 */

package net.citrusleaf;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.atomic.AtomicInteger;

// Node representing a single CitrusLeaf node.
public class CLNode
{
	public String name = "";    // unique name

	public InetSocketAddress primaryAddress; // the last place we successfully found this host - resolved

	public Vector<CLNodeAddress> aliases;  // host:port aliases for this node - may include RRdns entries and worse
	
	private static final int FULL_HAPPINESS = 300;
	private static final int MAX_TIME_SINCE_LAST_USED = 10000;
	
	// Something here about the keys that it likes to get
	public AtomicInteger happinessValue = new AtomicInteger(FULL_HAPPINESS);    // happiness value from 0 to 100.  Node starts fully happy.
	public int partition_generation = -1;
	
	public ArrayBlockingQueue<CLConnection> connQ;
	
	public int id;
	private static AtomicInteger id_counter = new AtomicInteger(); 
	
	protected CLConnectionManager myManager = null;
	
	static final int MAX_CONNECTIONS_PER_NODE = 300;
	
	public CLNode(CLConnectionManager myManager) {
		
		this.name   = "";
		this.aliases  = new Vector<CLNodeAddress>();
		
		this.connQ    = new ArrayBlockingQueue<CLConnection>(MAX_CONNECTIONS_PER_NODE);
		
		this.myManager = myManager;
		this.id = id_counter.getAndIncrement();
		
		this.primaryAddress = null;		
	}

	
	public CLNode(CLConnectionManager myManager, String hostname, int port) {
		
		this.name   = "";
		this.aliases  = new Vector<CLNodeAddress>();
		this.aliases.add(new CLNodeAddress(hostname, port));
		
		this.connQ    = new ArrayBlockingQueue<CLConnection>(MAX_CONNECTIONS_PER_NODE);
		
		this.myManager = myManager;
		this.id = id_counter.getAndIncrement();
		
		this.primaryAddress = null;		
	}
	
	// strips most of the interesting parts off the 
	public CLNode clone_node() {
		CLNode c = new CLNode(this.myManager);
		c.aliases = new Vector<CLNodeAddress>();
		for (CLNodeAddress na : this.aliases)
			c.aliases.add(na);
		return(c);
	}

	// the node's name never should change, except the first time, and
	// needs to be reflected in exterior pointers
	private boolean setName(String name) {

		if (this.name.equals("") == false) {
			CitrusleafClient.warn(" internal error: trying to set node "+this.name+" to "+name);
			return(false);
		}
		boolean rv = myManager.associateNodeName(this,name);
		if (rv==true) this.name = name;
		return(rv);
	}
	
	// Is this node - as identified by its NAME - in the list?
	public static CLNode nodeInList(ArrayList<CLNode> list, String name)
	{
		for (CLNode node : list) {
			if( node.name.equals(name) ){
				return node;
			}
		}
		return null;
	}
	
	// This is only used to make sure we don't keep adding new name/aliases when we already have a particular
	// one
	public static CLNode nodeInAliasList(ArrayList<CLNode> list, String hostname, int port)
	{		
		for (CLNode node : list) {
			for (CLNodeAddress alias : node.aliases) {
				if( alias.isEqual(hostname, port)){
					return node;
				}
			}
		}
		return null;
	}
	
	public static void nodeListDump(ArrayList<CLNode> list, String info) {
		if (! CitrusleafClient.debugEnabled()) {
			return;
		}
		
		CitrusleafClient.debug(info + " dump node list : size "+list.size());
		
		int i=0;
		for (CLNode node : list) {
			CitrusleafClient.debug(info + " NODE "+i+" name "+node.name);
			
			if (node.primaryAddress != null)
				CitrusleafClient.debug(i+" "+node.toString()+"   primary address "+ node.primaryAddress.toString() );
			else
				CitrusleafClient.debug(i+" "+node.toString()+"   no primary address " );
			
			for (CLNodeAddress alias : node.aliases) {
				CitrusleafClient.debug("   Alias " + alias.toString() );
			}
			i++;
		}
	}
	
	// add the aliases from this node into the other node, because they
	// are actually the same. This needs to be a unique add.
	//
	public void addAliases(CLNode newNode)
	{
		for (CLNodeAddress newAddress : newNode.aliases) {
			boolean found = false;
			CLNodeAddress address = null;
			Iterator<CLNodeAddress> i = this.aliases.iterator();
			while (i.hasNext()) {
				address = i.next();
				if (address.equals(newAddress)==true) {
					found = true;
					break;
				}
			}
			if (found == false && address != null) 	{
				aliases.add(newAddress);
			}
		}
	}
	
	public Boolean isEqual(String name)
	{
		if( name != null )
			return name.equals(name);
		else
			return false;
	}
	
	public String toString()
	{
		StringBuilder str = new StringBuilder();
		str.append("Name: ");
		str.append(name);
		if (primaryAddress != null) {
			str.append(" address:");
			str.append(primaryAddress.toString() );
		}
		else {
			str.append(" no_address");
		}

		// list of aliases:
		str.append(" Aliases: ");
		Iterator<CLNodeAddress> i = aliases.iterator();

		while(i.hasNext()){
			CLNodeAddress t = i.next();
			str.append(t.fullName);
			str.append(" ");
		}

		return str.toString();
	}

	public CLConnection getConnection() throws IOException, UnknownHostException
	{
		// If we're not properly connected to this node, return null immediately
		if( happinessValue.get() < 0 ){
			if (CitrusleafClient.debugEnabled()) {
				CitrusleafClient.debug("get connection: happiness 0: refusing to connect "+this.name);
			}
			return null;
		}
		
		CLConnection conn;

		conn = this.connQ.poll();
		
		while (conn != null) {			
			long timeSinceLastUsed = (new java.util.Date()).getTime() - conn.lastUsed.getTime();
			if(timeSinceLastUsed < MAX_TIME_SINCE_LAST_USED) {
				return(conn);
			} else {
				try {
					conn.close();
				} catch (Exception e) {
					if (CitrusleafClient.debugEnabled()) {
						CitrusleafClient.debug("cannot close old connection - server must have closed it for us");
					}
				}
			}
			conn = this.connQ.poll();
		}
		
 	    // None in queue - create!
		try {
			conn = new CLConnection (primaryAddress);
			conn.node = this;			
		} catch (IOException e) {
			if (CitrusleafClient.infoEnabled()) {
				CitrusleafClient.info(" IO exception connecting to "+primaryAddress.toString());
			}
			throw(e);
		}
		return(conn);
	}

	public void returnConnectionToPool(CLConnection conn)
	{		
	  	// interesting - if the app has more than this number outstanding, we'll thrash pretty hard
	  		
		// offer to the queue. If queue doesn't like it, just ignore and let the
		//reference counter sweep it up later
		if (false == connQ.offer(conn)) {
			conn.close();
		}
	}

	public void closeConnection(CLConnection conn)
	{
		conn.close();
	}
	
	public void closeAllConnections()
	{
		CLConnection conn = null;
		do {
		   conn = connQ.poll();
		   if ((conn != null) && (conn.isConnected() == true)) {
			   conn.close();
		   }
	   } while (conn != null);
	}
	
	// called by external users of this node when there is a valid data transfer. Sets the
	// happiness value back up.
	public void validNetworkActivity()
	{
        int hv = happinessValue.get();
        if (hv != FULL_HAPPINESS) {
        	if (CitrusleafClient.debugEnabled()) {
        		CitrusleafClient.debug("node "+name+" has valid network activity but not fully happy : "+hv);
        	}
        }
		happinessValue.set(FULL_HAPPINESS);
	}

	public void networkFailure()
	{
		int hv = happinessValue.addAndGet(-60);
		if (hv < 0) {
        	if (CitrusleafClient.debugEnabled()) {
        		CitrusleafClient.debug("node "+name+" general network failure, happiness decreased to "+hv+" dunning ");
        	}			
        	happinessValue.set(0);
			myManager.dunNode(this);
		}
        else {
        	if (CitrusleafClient.debugEnabled()) {
        		CitrusleafClient.debug("node "+name+" general network failure, happiness decreased to "+hv+" but still better than 0 ");
        	}
        }
	}

	private boolean getPrimaryAddress() {
		
		// for everything in the alias list, do a quick name check
		for (CLNodeAddress alias : aliases) {
			
			// get the list of addresses
			InetAddress[] addresses = null;
			try {
				addresses = InetAddress.getAllByName(alias.hostname);
			}
			catch (Exception e) {
				CitrusleafClient.warn("getPrimaryAddress: DNS failure for hostname "+alias.hostname+" exception: "+e.toString());
			}
			
			if (addresses == null || addresses.length == 0) {
				CitrusleafClient.warn("getPrimaryAddress: DNS failure for hostname "+alias.hostname);
				continue;
			}
				
			// for each address, see if you can learn something
			for (InetAddress address : addresses) {
				
				InetSocketAddress possibleAddress = new InetSocketAddress(address, alias.port);
			
				String possibleName = CitrusleafInfo.get(possibleAddress, "node");

				// got a real response					
				if (possibleName != null) {

					if (name.equals("") ) {						
						// I am unnamed. Name me. Side effect: may find out the node is already
						// named, so the name won't be set.
						// no? NAME ME
						//    a little care must be taken ...
						boolean rv = setName(possibleName);
						if (rv == true) primaryAddress = possibleAddress;
						return(rv);
					}
					else if (name.equals(possibleName)) {
						primaryAddress = possibleAddress;
						return(true);
					}
					else {
						// CitrusleafClient.info("getPrimaryAddress: found new node at adress in my alias list, what to do?");
						// todo: if this address was a direct IP address, forget it. If it was a round-robbin, maybe that's OK - should I start a new node ccreation or something?
					}
				}	
			}
		}
		return(false);
	}	
	
	// I think this means these will just hang out in global space and never be re-created
	static final String[] update_partition_services_strings = { "node", "partition-generation", "services" };
	static final String[] update_partition_strings = { "node", "partition-generation"  };
	static final String[] update_services_strings = { "node", "services" };
	static final String[] update_justnode_strings = { "node" };
	//static final String[] update_replicas_strings = { "replicas-write", "replicas-read" };
	static final String[] update_replicas_strings = { "replicas-write" };

	// synchronously requests canonical name for this node, and list of 
	// neighbor (ie 'friend') nodes. 
	//
	// Will create new CLNodes and call requestNodeInfo on those nodes as well.
	//
	// If you don't want to know about the friend nodes, don't pass in a list of friends
	// to be added to
	public void refreshNodeInfo(ArrayList<CLNode> possibleNewNodes, boolean update_partition) 
	{
		// if we don't yet have an address, go through the aliases and see what we can find
		if (primaryAddress == null) {
			
			// SIDE EFFECT: MIGHT SET NAME
			if (false == getPrimaryAddress()) {
				if (CitrusleafClient.infoEnabled()) {
					CitrusleafClient.info("requestNodeInfo: "+name+" could not get primary address");
				}
				return;
			}
			// all they really wanted was updating the name, so just return - fast path
			if (possibleNewNodes == null && update_partition == false) 
				return;
		}		
		
		if (primaryAddress == null) {
			if (CitrusleafClient.debugEnabled()) {
				CitrusleafClient.debug("requestNodeInfo: "+name+" no primary address");
			}
			return;
		}
		
		String[] reqStrings;
		if (update_partition == true) {
			if (possibleNewNodes == null)   reqStrings = update_partition_strings;
			else 							reqStrings = update_partition_services_strings;
		}
		else { // no partition
			if (possibleNewNodes == null)   reqStrings = update_justnode_strings;
			else							reqStrings = update_services_strings;
		}

		HashMap <String, String> nodeInfo = CitrusleafInfo.get(primaryAddress, reqStrings);
		if ( nodeInfo == null ) {
			
			if (CitrusleafClient.debugEnabled()) {
				CitrusleafClient.debug("node "+name+" info get failed, decrease happiness "+primaryAddress.toString());
			}
			
			int hv = happinessValue.addAndGet(-50);
			if (hv < 0) {
			
				if (CitrusleafClient.debugEnabled()) {
					CitrusleafClient.debug("node "+name+" info get failed, happiness decreased to "+hv+" dunning "+primaryAddress.toString());
				}
				happinessValue.set(0);
				
				// clean this node out of upper level tables
				myManager.dunNode(this);
				partition_generation = -1;
			}
			return;
		}
			
		String info_name = nodeInfo.get("node");
		String friends  = nodeInfo.get("services");

		if (info_name == null || info_name.equals("")) {
			if (CitrusleafClient.debugEnabled()) {
				CitrusleafClient.debug("request to hostname "+primaryAddress.toString()+" returned no node name.");
			}
			return;
		}
		
		// check to see this is the node I'm expecting. 
		// in the case of round-robbin dns, I might have the hostname that resolves
		// to a new address.
		// in any case, the primary address no longer points to this node, so blow that away, and hope
		// one of the aliases saves us later
		if (name.equals(info_name) == false) {
			
			if (CitrusleafClient.debugEnabled()) {
				CitrusleafClient.debug("request to hostname "+primaryAddress.toString()+" returned node name "+info_name+" instead of "+name);
			}
			// todo: can I route this response to the correct node, if I already know about it?
			primaryAddress = null;			
			return;
		}
		
		validNetworkActivity();

		// parse the service addresses, add the friends to the list, but don't get
		// their info - another pass through the list will do so
		if ((friends != null) && (possibleNewNodes != null)) {  // if single node, no friends. Sad!
			
			// CitrusleafClient.info(" requestNodeInfo: "+id+" services: " + friends);
			String friendNames[] = friends.split(";");

			for( int i=0; i<friendNames.length; i++ ) {
				String friend = friendNames[i];
				try {

					String friendInfo[] = friend.split(":");
					String friendHostname = friendInfo[0];
					int friendPort = Integer.parseInt(friendInfo[1]);
//					CitrusleafClient.info(" Check found possible service " + friendHostname + ":"+ friendPort);
					CLNode newNode = null;
					synchronized (myManager.nodesArray) {
						if( (nodeInAliasList(myManager.nodesArray, friendHostname, friendPort) == null) &&
							(nodeInAliasList(possibleNewNodes, friendHostname, friendPort) == null) ) {
						
							// CitrusleafClient.info(" request info found FRIENDLY node, creating "+friendHostname+":"+friendPort);
							// nodeListDump(myManager.nodesArray, " Manager node list: ADDING FRIEND :");
							newNode = new CLNode(myManager, friendHostname, friendPort);
							possibleNewNodes.add(newNode);
						}
					}
				} catch(Exception e) {				
				}
			}
		}

		if (update_partition) {
			
			String genStr = nodeInfo.get("partition-generation");
			int new_partition_generation = -1;
			if (genStr != null) new_partition_generation = Integer.parseInt( genStr );

			if (new_partition_generation != partition_generation) {
				// arguably, we should remove the association as soon as we know something. But, realistically,
				// we don't want to have the transactions routed poorly while we're doing the request				
				
				// request up-to-date
				HashMap<String,String> resp = CitrusleafInfo.get(primaryAddress, update_replicas_strings);
				
				String replicasList = resp.get("replicas-write"); 
				if (replicasList != null) {
					// Do not clean this node out of upper level tables.
					// The location of the other node will eventually be replaced by the other node refreshes.
					// myManager.unassociateNodePartitionAll(this, CLConnectionManager.PartitionType.WRITE);
					
					// parse and insert
					for (String repl : replicasList.split(";")) {
						myManager.associateNodePartition(this, repl, CLConnectionManager.PartitionType.WRITE);
					}
					partition_generation = new_partition_generation;
				}
				else {
					if (CitrusleafClient.infoEnabled()) {
						CitrusleafClient.info(" could not fetch write replicas from node "+this.name);
					}
				}

				/* The read partition hashmap is never used, so why bother updating it?
				replicasList = resp.get("replicas-read"); 
				if (replicasList != null) {
					// clean this node out of upper level tables
					myManager.unassociateNodePartitionAll(this, CLConnectionManager.PartitionType.READ);
					
					// parse and insert
					// name\tvalue\n
					for (String replica : replicasList.split("\n")) {
						// CitrusleafClient.info(" node "+this.name+" responsible for "+replica);
						for (String repl : replica.split(";")) {
							myManager.associateNodePartition(this, repl, CLConnectionManager.PartitionType.READ);
						}
					}
					partition_generation = new_partition_generation;

				}
				else {
					if (CitrusleafClient.infoEnabled()) {
						CitrusleafClient.info(" could not fetch write replicas from node "+this.name);
					}
				}
				*/				
			}
		}
	}	
}
