/*
 * Citrusleaf Aerospike Client - Java Library
 *
 * Copyright 2009-2010 by Citrusleaf, Inc. All rights reserved.
 *
 * Availability of this source code to partners and customers includes
 * redistribution rights covered by individual contract. Please check your
 * contract for exact rights and responsibilities.
 */

package net.citrusleaf;

import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.util.HashMap;

import net.citrusleaf.CitrusleafClient.SerializeException;

/**
 * Access Citrusleaf's monitoring protocol - the "Info" protocol.
 * <p>
 * The info protocol is a name/value pair based system, where an individual
 * database server node is queried to determine its configuration and status.
 * The list of supported names can be found on the Citrusleaf Wiki under the TCP
 * wire protocol specification.
 * <p>
 * Citrusleaf info values are accessible via a node's host name and port, or
 * directly via its <code>InetSocketAddress</code>.
 */
public class CitrusleafInfo {

	//=================================================================
	// Constants
	//

	private static final int DEFAULT_TIMEOUT = 2000;


	//=================================================================
	// Public Functions
	//

	//-------------------------------------------------------
	// Get Info via Host Name and Port
	//-------------------------------------------------------

	/**
	 * Get all the default info from the specified database server node, using
	 * host name and port.
	 * 
	 * @param hostname			host name
	 * @param port				host port
	 * @return					info name/value pairs
	 */
	public static HashMap<String, String> get(String hostname, int port) {
		InetSocketAddress isa = null;

		try {
			isa = new InetSocketAddress(hostname, port);
		}
		catch (Exception e) {
		}

		return isa != null ? get(isa) : null;
	}

	/**
	 * Get one info value by name from the specified database server node, using
	 * host name and port.
	 * 
	 * @param hostname			host name
	 * @param port				host port
	 * @param name				name of value to retrieve
	 * @return					info value
	 */
	public static String get(String hostname, int port, String name) {
		InetSocketAddress isa = null;

		try {
			isa = new InetSocketAddress(hostname, port);
		}
		catch (Exception e) {
		}

		return isa != null ? get(isa, name) : null;
	}

	/**
	 * Get many info values by name from the specified database server node,
	 * using host name and port.
	 * 
	 * @param hostname			host name
	 * @param port				host port
	 * @param names				names of values to retrieve
	 * @return					info name/value pairs
	 */
	public static HashMap<String, String> get(String hostname, int port,
			String[] names) {
		InetSocketAddress isa = null;

		try {
			isa = new InetSocketAddress(hostname, port);
		}
		catch (Exception e) {
		}

		return isa != null ? get(isa, names) : null;
	}

	//-------------------------------------------------------
	// Get Info via Socket Address
	//-------------------------------------------------------

	/**
	 * Get all the default info from the specified database server node.
	 * 
	 * @param isa				<code>InetSocketAddress</code> of server node
	 * @return					info name/value pairs
	 */
	public static HashMap<String, String> get(InetSocketAddress isa) {
		Buf req_buf = prepare(null);
		String response = request(isa, req_buf, DEFAULT_TIMEOUT);

		return parseResponse(response);
	}

	/**
	 * Get one info value by name from the specified database server node.
	 * 
	 * @param isa				<code>InetSocketAddress</code> of server node
	 * @param name				name of value to retrieve
	 * @return					info value
	 */
	public static String get(InetSocketAddress isa, String name) {
		Buf req_buf = prepare(name + "\n");
		String response = request(isa, req_buf, DEFAULT_TIMEOUT);
		HashMap<String, String> responses = parseResponse(response);

		return responses != null ? responses.get(name) : null;
	}

	/**
	 * Get many info values by name from the specified database server node.
	 * 
	 * @param isa				<code>InetSocketAddress</code> of server node
	 * @param names				names of values to retrieve
	 * @return					info name/value pairs
	 */
	public static HashMap<String, String> get(InetSocketAddress isa,
			String[] names) {
		StringBuilder reqStr = new StringBuilder();

		for (String name : names) {
			reqStr.append(name);
			reqStr.append('\n');
		}

		Buf req_buf = prepare(reqStr.toString());
		String response = request(isa, req_buf, DEFAULT_TIMEOUT);

		return parseResponse(response);
	}


	//=================================================================
	// Private Functions
	//

/* Not used, but leaving it here in case anyone wants to revive it. [AKG]

	private static class InfoResponse {
		InetSocketAddress responseAddr;
		String response;
	}

	// In the rare cases where it is permissible to make a request to a host
	// name, use this helper function, and perhaps pay attention to the
	// returned socket address.
	private static InfoResponse request(String host, int port, Buf buf) {
		InetSocketAddress isa;

		try {
			isa = new InetSocketAddress(InetAddress.getByName(host), port);
		}
		catch (UnknownHostException e) {
			CitrusleafClient.info(" info: could not resolve hostname " + host +
					":" + port);
			return null;
		}

		InfoResponse rv = new InfoResponse();

		rv.responseAddr = isa;
		rv.response = request(isa, buf, DEFAULT_TIMEOUT);

		return rv;
	}
*/

	// For most requests, its important to know the exact socket address where
	// the request was made. Also, timeouts of the DNS system must be managed
	// carefully. For these reasons, force the upper layer to pass in an entire
	// InetSocketAddress thus the resolution has already been done.
	private static String request(InetSocketAddress isa, Buf buf, int timeout) {
		// Connect.
		Socket s = null;

		try {
			s = new Socket();
			s.connect(isa, timeout);
			s.setTcpNoDelay(true);
			s.setSoTimeout(timeout); // not great, but should be very close -
									 // we're unlikely to have lots of timeouts
		}
		catch (Exception e) {
			// This happens. It's OK, because sometimes we keep pinging nodes
			// that are down.
			if (CitrusleafClient.infoEnabled()) {
				CitrusleafClient.info(" info: could not connect to "+
						isa.getHostName() + " : " + e.getMessage());
			}

			if (s != null) {
				try {
					s.close();
				}
				catch (Exception e2) {
				}
			}

			return null;
		}

		// Write.
		OutputStream os = null;

		try {
			os = s.getOutputStream();
			os.write(buf.buf, 0, buf.len);
		}
		catch (Exception e) {
			if (CitrusleafClient.infoEnabled()) {
				CitrusleafClient.info(" info: write failed: " + e.getMessage());
			}

			if (s != null) {
				try {
					s.close();
				}
				catch (Exception e2) {
				}
			}

			return null;
		}

		// Read - Let's reuse the input buffer.
		InputStream is;

		try {
			is = s.getInputStream();

			int off = 0;

			while (off < 8) {
				int r_b = is.read(buf.buf, off, 8 - off);

				if (r_b < 0) {
					return null;
				}

				off += r_b;
			}

			long szField = CLBuffer.get_ntohll(buf.buf, 0);

			buf.len = (int)(szField & 0xFFFFFFFFFFFFL);

			if (buf.len > 0) {
				if (buf.len > buf.allocLen) {
					buf.buf = new byte[buf.len];
					buf.allocLen = buf.len;
				}

				off = 0;

				while (off < buf.len) {
					int r_b = is.read(buf.buf, off, buf.len - off);

					if (r_b < 0) {
						return null;
					}

					off += r_b;
				}
			}
		}
		catch (Exception e) {
			// Catch here means the server gave up on us or something.
			if (CitrusleafClient.infoEnabled()) {
				CitrusleafClient.info(" info: io failed: " + e.getMessage() +
						" " + e.toString());
			}

			if (s != null) {
				try {
					s.close();
				}
				catch (Exception e2) {
				}

				s = null;
			}

			return null;
		}

		// Close.
		try {
			s.close();
		}
		catch (Exception e) {
		}

		// Convert the UTF8 byte array into a string.
		if (buf.len > 0) {
			try {
				return CLBuffer.utf8ToString(buf.buf, 0, buf.len);
			}
			catch (SerializeException e) {
				return null;
			}
		}	
		else {
			if (CitrusleafClient.infoEnabled()) {
				CitrusleafClient.info(" info: can't convert utf8 to unicode");
			}

			return null;
		}
	}

	private static Buf prepare(String req) {
		byte msgBuf[];
		int sz;
		Buf rv = new Buf();

		// Remember that UTF8 can explode unicode out a bit, so over-allocate
		// the buffer and be sure to use the actual size returned.
		if (req != null) {
			rv.allocLen = (req.length() * 4) + 8;
			msgBuf = new byte[rv.allocLen];		
			sz = CLBuffer.stringToUtf8(req, msgBuf, 8);
		}
		else {
			rv.allocLen = 0;
			sz = 0;
			msgBuf = new byte[8];
		}

		long sz_field = (sz) | (2L << 56) | (1L << 48);

		CLBuffer.set_htonll(sz_field, msgBuf, 0);
		rv.buf = msgBuf;
		rv.len = sz + 8;

		return rv;
	}

	// Parse out a response string.
	private static HashMap<String,String> parseResponse(String response) {
		if (response == null) {
			return null;
		}

		HashMap<String, String> responses = new HashMap<String, String>();
		String lines[] = response.split("\n");

		for (String line : lines) {
			String nv[] = line.split("\t");

			if (nv.length >= 2) {
				responses.put(nv[0], nv[1]);
			}
			else if (nv.length == 1) {
				responses.put(nv[0], null);
			}
		}

		return responses.size() != 0 ? responses : null;
	}


	//=================================================================
	// Private Nested Interfaces & Classes
	//

	private static class Buf {
		int len;
		int allocLen;
		byte[] buf;
	}
}
