/*
 *  Citrusleaf client examples
 *
 *  Copyright 2011 by Citrusleaf, Inc.  All rights reserved.
 *  
 *  Availability of this source code to partners and customers includes
 *  redistribution rights covered by individual contract. Please check
 *  your contract for exact rights and responsibilities.
 */

/*
 * Use notes. This stand-alone application provides a number of functional
 * end user tests. Simply compile this file with the Citrusleaf library
 * and enjoy the fine example-ish goodness. 
 * 
 * It does rely on the Apache Commons CLI (Command Line Interpreter)
 * library available from http://commons.apache.org/cli/ . This library has been stable
 * in its 1.2 version since 2002.
 */

package net.citrusleaf.examples;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.reflect.Constructor;

import javax.swing.JPanel;

import net.citrusleaf.CitrusleafClient;
import net.citrusleaf.CitrusleafClient.ClResultCode;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;

public class Main extends JPanel {

	private static final long serialVersionUID = 1L;
	private static final String[] ExampleNames = new String[] {
		"ServerInfo",
		"SetGet",
		"Add",
		"Append",
		"Prepend",
		"BatchGet",
		"Generation",
		"Serialize",
		"Expire",
		"Touch",
		"DeleteBin",
		"ScanParallel",
		"ScanSeries"
	};
	public static String[] getAllExampleNames() { return ExampleNames; }

	/**
	 * Main entry point.
	 */
	public static void main(String[] args) {

		Console console = new Console();

		try {
			Options options = new Options();
			options.addOption("h", "host", true, "Server hostname (default: localhost)");
			options.addOption("p", "port", true, "Server port (default: 3000)");
			options.addOption("n", "namespace", true, "Namespace (default: test)");
			options.addOption("s", "set", true, "Set name (default: emptystring)");
			options.addOption("g", "gui", false, "invoke GUI to selectively run tests");
			options.addOption("u", "usage", false, "Print usage.");

			CommandLineParser clp = new GnuParser();
			CommandLine cl = clp.parse(options, args, false);

			if (args.length == 0 || cl.hasOption("u")) {
				logUsage(console, options);
				return;
			}
			Parameters params = parseParameters(cl);
			String[] exampleNames = cl.getArgs();
			
			if ((exampleNames.length == 0) && (!cl.hasOption("g"))) {
				logUsage(console, options);
				return;			
			}
			processExamples(console, params, exampleNames, cl.hasOption("g"));
		}
		catch (Exception ex) {
			console.error(ex.getMessage());
			ex.printStackTrace();
		}
	}

	/**
	 * Write usage to console.
	 */
	private static void logUsage(Console console, Options options) {
		HelpFormatter formatter = new HelpFormatter();
		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);
		String syntax = Main.class.getName() + " [<options>] all|(<example1> <example2> ...)";
		formatter.printHelp(pw, 100, syntax, "options:", options, 0, 2, null);
		console.log(sw.toString());
		console.log("examples:");

		for (String name : ExampleNames) {
			console.log(name.toString());			
		}
		console.log("");
		console.log("All examples will be run if 'all' is specified as an example.");
	}

	/**
	 * Parse command line parameters.
	 */
	private static Parameters parseParameters(CommandLine cl) throws Exception {
		String host = cl.getOptionValue("h", "127.0.0.1");
		String portString = cl.getOptionValue("p", "3000");
		int port = Integer.parseInt(portString);
		String namespace = cl.getOptionValue("n","test");
		String set = cl.getOptionValue("s", "");

		Parameters params = new Parameters(host, port, namespace, set);
		params.setIsSingleBin();
		return params;
	}
	
	/**
	 * Connect and run one or more client examples.
	 */
	private static void processExamples(Console console, Parameters params, String[] examples, boolean invoke_gui) throws Exception {
		CitrusleafClient cs = new CitrusleafClient();

		try {
			if (cs.addHost(params.host, params.port) != ClResultCode.OK) {
				throw new Exception(String.format("Failed to add %s:%d to cluster.", 
					params.host, params.port));
			}

			// Check for all.
			for (String exampleName : examples) {
				if (exampleName.equalsIgnoreCase("all")) {
					examples = ExampleNames;
					break;
				}
			}

			if (invoke_gui) {
				GuiDisplay.startGui(examples, cs, params, console);
			}
			else {
				for (String exampleName : examples) {
					runExample(exampleName, cs, params, console);
				}
			}
		}
		finally {
			cs.close();
		}
	}

	/**
	 * Run client example.
	 */
	public static void runExample(String exampleName, CitrusleafClient cs, Parameters params, Console console) throws Exception {
		String fullName = "net.citrusleaf.examples." + exampleName;
		Class<?> cls = Class.forName(fullName);

		if (Example.class.isAssignableFrom(cls)) {
			Constructor<?> ctor = cls.getDeclaredConstructor(Console.class);
			Example example = (Example)ctor.newInstance(console);
			example.run(cs, params);
		}
		else {
			console.error("Invalid example: " + exampleName);
		}
	}
}
